

!!! Listing 1

%token Number

%left   '+' '-'
%left   '*' '/'
%right  UNARY

%start  prog

%%

expr    : expr '+' expr
    | expr '-' expr
    | expr '*' expr
    | expr '/' expr
    | '+' expr %prec UNARY
    | '-' expr %prec UNARY
    | '(' expr ')'
    | Number

prog    : /* null */
    | prog expr '\n'
    | prog '\n'

%%

!!! Listing 2

%{
package arith_simple;

import java.io.*;
import java.util.*;

public class Arith {
%}

%token  <Double>    Number 99
%type   <Double>    expr

%left   '+' '-'
%left   '*' '/'
%right  UNARY

%start  prog

%%  // public Object yyparse(yyInput yyLex) throws IOException, yyException

expr    : expr '+' expr { $$ = new Double($1.doubleValue()+$3.doubleValue()); }
    | expr '-' expr     { $$ = new Double($1.doubleValue()-$3.doubleValue()); }
    | expr '*' expr     { $$ = new Double($1.doubleValue()* $3.doubleValue()); }
    | expr '/' expr     { $$ = new Double($1.doubleValue()/$3.doubleValue()); }
    | '+' expr %prec UNARY  { $$ = $<>2; }  // can suppress class
    | '-' expr %prec UNARY  { $$ = new Double(-$2.doubleValue()); }
    | '(' expr ')'      { $$ = $2; }
    | Number            // $$ = yyDefault($1);

prog    : /* null */
    | prog expr '\n'    { System.out.println("\t"+$2); }
    | prog '\n'

%%

  ..... siehe unten
}

!!! Listing 3

class Scanner extends StreamTokenizer implements Arith.yyInput {
  public Scanner (Reader r) {
    super (r);
    eolIsSignificant(true);     // need '\n'
    ordinaryChar('/');          // gotcha: would start comment
    ordinaryChar('-');          // gotcha: would start Number
    commentChar('#');           // comments from # to end-of-line
  }

  public boolean advance () throws IOException {
    return ttype != TT_EOF && super.nextToken() != TT_EOF;
  }

  public int token () {
    value = null;
    switch (ttype) {
    case TT_EOF:    return 0;   // should not happen
    case TT_EOL:    return '\n';
    case TT_NUMBER: value = new Double(nval);
                    return Arith.Number;
    case TT_WORD:   return Arith.yyErrorCode;
    default:        return ttype;
    }
  }

  protected Object value;

  public Object value () {
    return value;
  }
}

!!! Listing 4

%%
...
%%

  public static void main (String args []) {
    Scanner scanner = new Scanner(new InputStreamReader(System.in));
    try {
      new Arith().yyparse(scanner);
    } catch (IOException ie) { ie.printStackTrace(); }
      catch (yyException ye) { System.err.println(ye); }
  }
}

!!! Listing 5

import java.io.Serializable;

public abstract class Node extends Number implements Serializable {

  protected abstract static class Binary extends Node {
    protected Number left;
    protected Number right;

    protected Binary (Number left, Number right) {
      this.left = left; this.right = right;
    }
  }

  public static class Add extends Binary {
    public Add (Number left, Number right) {
      super(left, right);
    }
    ...
    public double doubleValue () {
      return left.doubleValue() + right.doubleValue();
    }
  }
  ...
}

!!! Listing 6

%{
    ...
%}

%token  <Number>    Constant
%type   <Number>    sum, product, term
%type   <Vector>    line

%%

line    : sum '\n'      { $$ = new Vector(); $<Vector>$.addElement($1); }
    | line sum '\n'     { $1.addElement($2); }

sum : product           // $$ = $1
    | sum '+' product   { $$ = new Node.Add($1, $3); }
    | sum '-' product   { $$ = new Node.Sub($1, $3); }

product : term          // $$ = $1
    | product '*' term  { $$ = new Node.Mul($1, $3); }
    | product '/' term  { $$ = new Node.Div($1, $3); }
    | product '%' term  { $$ = new Node.Mod($1, $3); }

term    : '+' term      { $$ = $2; }
    | '-' term          { $$ = new Node.Minus($2); }
    | '(' sum ')'       { $$ = $2; }
    | Constant          // $$ = $1
%%
}

!!! Listing 7

%{
  ...
  public class Arith {

    public static void main (String args []) {
      boolean cflag = args.length > 0 && args[0].equals("-c");
      Scanner scanner = new Scanner(new InputStreamReader(System.in));
      Arith parser = new Arith();

      try {
        Vector vec = (Vector)parser.yyparse(scanner);
        if (vec != null)
          if (cflag) {
            ObjectOutputStream out = new ObjectOutputStream(System.out);
            out.writeObject(vec);
            out.close();
          } else
            for (Enumeration e = vec.elements() ; e.hasMoreElements() ;) {
              System.out.println(((Number) e.nextElement()).floatValue());
            }
      } catch (yyException ye) { System.err.println(scanner+": "+ye); }
        catch (IOException ioe) { System.err.println(ioe); }
    }
%}

!!! Listing 8

import java.io.ObjectInputStream;
import java.util.Vector;

public class Go {
  public static void main (String args []) {
    try {
      ObjectInputStream in = new ObjectInputStream(System.in);
      Vector lines = (Vector)in.readObject();
      System.out.println("long\tdouble");
      for (int i = 0; i < lines.size(); ++ i) {
        Number n = (Number)lines.elementAt(i);
        System.out.println(n.longValue()+"\t"+n.doubleValue());
      }
    } catch (Exception e) { System.err.println(e); }
  }
}

!!! Listing 9

$ make test
{ echo 2+3;  echo '32768 * 3.2 + 128'; \
    } | CLASSPATH=.. java arith_tree.Arith
5.0
104985.6
{ echo 2+3;  echo '32768 * 3.2 + 128'; \
    } | CLASSPATH=.. java arith_tree.Arith -c \
      | CLASSPATH=.. java arith_tree.Go
long    double
5       5.0
98432   104985.6

!!! Listing 10

$ echo 2+3 | CLASSPATH=.. java arith_debug.Arith trace
push	state 0	value null
lex	state 0	reading Constant	value 2
shift	from state 0 to 2
push	state 2	value 2
reduce	state 2	uncover 0	rule (15) term : Constant
goto	from state 0 to 9
push	state 9	value 2
reduce	state 9	uncover 0	rule (8) product : term
goto	from state 0 to 8
push	state 8	value 2
lex	state 8	reading '+'	value null
reduce	state 8	uncover 0	rule (5) sum : product
goto	from state 0 to 7
...
reduce	state 15	uncover 0	rule (1) line : sum '\n'
goto	from state 0 to 6
lex	state 6	reading end-of-file	value null
accept	value [arith_debug.Node$Add@80cd04e]
5.0

